"""image.py

This module represents the image object
"""

import os
from tempfile import mkstemp
from PIL import Image, ImageDraw, ImageFont

MODE = 'RGB'
WHITE = 'rgb(255,255,255)'
LIGHT_GREY = 'rgb(175,175,175)'
BOX_SIZE = 50
Y_TEXT_OFFSET = -2
NUMBER_OFFSET = 4
ANSWER_OFFSET = 13
FILE_PATH = os.path.dirname(os.path.realpath(__file__))
FONT_PATH = FILE_PATH+"/static/FreeSans.ttf"


class CrosswordImage(object):

    def __init__(self, n, m, board):
        self._board = board
        self._width = n*BOX_SIZE
        self._height = m*BOX_SIZE
        self._image = Image.new(MODE, (self._width+1, self._height+1),
                                color=WHITE)
        self._draw = ImageDraw.Draw(self._image)
        self._num_font = None
        self._text_font = None
        print FONT_PATH
        self._num_font = ImageFont.truetype(FONT_PATH, 
                                            size=ANSWER_OFFSET)
        self._text_font = ImageFont.truetype(FONT_PATH, 
                                             size=BOX_SIZE-int(1.75*ANSWER_OFFSET))

    def construct_image(self):
        self._create_grid()
        self._populate_grid()

    def _create_grid(self):
        """ Create lines for grid. """
        # Draw row
        for i in range(0, self._height+1, BOX_SIZE):
            self._draw.line(((0,i),(self._width,i)), fill=128)

        # Draw column
        for i in range(0, self._width+1, BOX_SIZE):
            self._draw.line(((i,0),(i,self._height)), fill=128)

    def _populate_grid(self):
        """ Populate grid with items. """
        x, y = 0, 0
        for row in self._board:
            for box in row:
                # Fill in blank boxes
                if box.is_blank:
                    corner1 = (x, y)
                    corner2 = (x+BOX_SIZE, y+BOX_SIZE)
                    self._draw.rectangle((corner1, corner2), fill=128)
                # Fill in empty boxes
                else:
                    # Fill in numbers
                    number = str(box.number) if box.number else ''
                    loc = (x+NUMBER_OFFSET, y+NUMBER_OFFSET+Y_TEXT_OFFSET)
                    self._draw.text(loc, number, fill=128, font=self._num_font)
                    # Fill in text
                    loc = (x+ANSWER_OFFSET, y+ANSWER_OFFSET+Y_TEXT_OFFSET)
                    if box.has_submission:
                        print 'submit: %s' % box.latest_submission
                        self._draw.text(loc, box.latest_submission, fill=128, 
                                        font=self._text_font)
                    elif box.has_ghost:
                        print 'ghost: %s' % box.latest_ghost
                        self._draw.text(loc, box.latest_ghost, fill=LIGHT_GREY,
                                        font=self._text_font)

                # Move to the next boxes
                x = (x + BOX_SIZE) % self._width
            y = y + BOX_SIZE

    def save(self):
        _handle, path = mkstemp()
        path += '.png'
        # self._image.save('testcw.png', 'PNG')
        self._image.save(path, 'PNG')
        return path




"""puzzle.py

This module represents the puzzle
"""

__title__ = 'puzcw'
__author__ = 'Brian Wu'
__author_email__ = 'brian.george.wu@gmail.com'

import urllib2
from datetime import datetime, timedelta
import math
from image import CrosswordImage
from exception import BoardDimensionException, \
    WriteException

from box import Box
from clues import Clue, Clues, ACROSS_ID, DOWN_ID

NULL = '\x00'
OPEN_BOX = '-'
BLANK = '.'
BOARD_START_KEY = OPEN_BOX*3
BOARD_IDX_KEY = 'NY Times'


class Puzzle(object):

    def __init__(self):
        # Initialize empty game
        self.width = None
        self.height = None
        self.board = []
        self.clues = Clues()

    def init_game(self, n=None, m=None):
        """ Retrieve and parse game. """
        self.__init__(n, m)
        game_str = self._download_game()
        self._process_game_str(game_str)
        self.save_game()

    def save_game(self):
        """ Save crossword as image, return path. """
        image = CrosswordImage(self.width, self.height, self.board)
        image.construct_image()
        return image.save()

    def get_clue(self, num, clue_type):
        """ Retrieve clue by key. """
        num = int(num)
        clue_type = clue_type.lower()
        try:
            return self.clues.get_clue(num, clue_type)
        except KeyError:
            return "Clue %s:%s doesn't exist!" % (num, clue_type)

    @property
    def across_clues(self):
        """ Retrieve all across clues. """
        return self.clues.clues_by_type(ACROSS_ID)

    @property
    def down_clues(self):
        """ Retrieve all down clues. """
        return self.clues.clues_by_type(DOWN_ID)

    @property
    def all_clues(self):
        return self.across_clues + self.down_clues

    def _write(self, num, clue_type, word, guess_type):
        clue = self.get_clue(num, clue_type)
        x, y = clue.start_position
        for letter in word:
            box = self.board[x][y]

            # Raise WriteException for blank box
            if box.is_blank: raise WriteException

            box.write(guess_type, clue_type, letter.upper())

            # Increment x or y
            if clue_type == ACROSS_ID:
                y += 1
            else:
                x += 1

    def submit(self, num, clue_type, word):
        self._write(num, clue_type, word, 'submission')

    def ghost(self, num, clue_type, word):
        self._write(num, clue_type, word, 'ghost')

    def clear(self, num, clue_type):
        """ Clear answers from boxes by direction. """
        clue = self.get_clue(num, clue_type)
        x, y = clue.start_position
        while True:
            try:
                box = self.board[x][y]
                if box.is_blank: raise WriteException
                box.clear(clue_type)
                if clue_type == ACROSS_ID:
                    y += 1
                else:
                    x += 1

            # Break at a blank or index error
            except (WriteException, IndexError) as e:
                break

    # def _download_game(self):
    #     """ Download and format game. """
    #     day = datetime.now()
    #     attempts = 0
    #     while attempts < 10:
    #         try:
    #             day_str = day.strftime('%y.%m.%d.puz')
    #             # day_str = '16.07.03.puz'
    #             url = MOTHERLOAD + day_str
    #             res = urllib2.urlopen(url)
    #             ucontent = unicode(res.read(), "ISO-8859-1")
    #             return ucontent.encode('utf-8')
    #         except urllib2.HTTPError:
    #             attempts += 1
    #             day = day - timedelta(days=1)

    @classmethod
    def from_url(cls, url):
        res = urllib2.urlopen(url)
        game_str = unicode(res.read(), "ISO-8859-1").encode('utf-8')
        return cls.from_str(game_str)

    @classmethod
    def from_str(cls, game_str):
        instance = cls()
        instance._build_board(game_str)
        return instance

    def _build_board(self, game_str):
        """ 
        Process the board string the board string. 

        1. Determine index of board start
        2. Determine n
        3. Create board and boxes
        """
        game_list = game_str.split(NULL)

        # Get board
        for idx, data in enumerate(game_list):
            if BOARD_IDX_KEY in data:
                board_idx = idx
                board_str = data
                break

        board_start = board_str.index(BOARD_START_KEY)
        sqrt = math.sqrt(board_start)
        is_square = sqrt - int(sqrt) == 0
        if is_square:
            self.width = int(math.sqrt(board_start))
            self.height = self.width
        elif self.width is None or self.height is None:
            msg = "Non square board detected."
            raise BoardDimensionException(msg)

        answer_idx = 0
        clue_count = 1
        clue_idx = board_idx + 3
        for row in range(self.height):
            new_row = []
            for col in range(self.width):

                # Determine info for box and clues
                board_space = board_str[board_start + answer_idx]
                is_blank = board_space==BLANK 
                answer = None if is_blank else board_str[answer_idx]
                is_across = not is_blank and (col==0 or new_row[col-1].is_blank)
                is_down = not is_blank and (row==0 or self.board[row-1][col].is_blank)
                if is_across or is_down:
                    number = clue_count
                    if is_across:
                        clue = Clue(clue=game_list[clue_idx],
                                    start_position=(row,col),
                                    num=number,
                                    clue_type=ACROSS_ID)
                        self.clues.put(clue_count, ACROSS_ID, clue)
                        clue_idx += 1
                    if is_down:
                        clue = Clue(clue=game_list[clue_idx],
                                    start_position=(row,col),
                                    num=number,
                                    clue_type=DOWN_ID)
                        self.clues.put(clue_count, DOWN_ID, clue)
                        clue_idx += 1
                    clue_count += 1
                else:
                    number = None

                # Create and append box
                box = Box(answer=answer,
                          is_blank=is_blank,
                          number=number)
                new_row.append(box)
                answer_idx += 1

            self.board.append(new_row)




"""clues.py

This module represents the clues for the puzzle object
"""

from time import time
from collections import namedtuple

AnswerTuple = namedtuple('AnswerTuple', 'val ts')


class Box(object):

    def __init__(self, answer, is_blank, number):
        self.answer = answer
        self.is_blank = is_blank
        self.number = number

        # Filled in later
        self._submission = {}
        self._ghost = {}

    def __str__(self):
        number = 0 if self.number is None else self.number
        answer = '_' if self.answer is None else self.answer
        return '%s:%s' % (number, answer)

    def clear(self, clue_type):
        """ Clear box object. """
        self._submission.pop(clue_type, None)
        self._ghost.pop(clue_type, None)

    def write(self, write_type, clue_type, letter):
        """ Write a letter with timestamp. """
        cache = getattr(self, '_%s' % write_type)
        cache[clue_type] = AnswerTuple(val=letter, ts=time())

    @property
    def has_submission(self):
        return len(self._submission) > 0

    @property
    def has_ghost(self):
        return len(self._ghost) > 0

    @property
    def latest_submission(self):
        latest = AnswerTuple(val='', ts=0)
        for k, v in self._submission.iteritems():
            if v.ts > latest.ts:
                latest = v
        return latest.val

    @property
    def latest_ghost(self):
        latest = AnswerTuple(val='', ts=0)
        for k, v in self._ghost.iteritems():
            if v.ts > latest.ts:
                latest = v
        return latest.val


"""clues.py

This module represents the clues for the puzzle object
"""

from collections import namedtuple

ACROSS_ID = 'a'
DOWN_ID = 'd'

ClueKey = namedtuple('ClueKey', ['num', 'clue_type'])


class Clue(object):

    def __init__(self, clue, start_position, num, clue_type):
        self.clue = clue
        self.start_position = start_position
        self.num = num
        self.clue_type = clue_type

    def __repr__(self):
        return '%s:%s:%s' % (self.clue, self.num, self.clue_type)


class Clues(object):

    def __init__(self):
        self.clues = {}

    def put(self, num, clue_type, clue):
        k = ClueKey(num, clue_type)
        self.clues[k] = clue

    @property
    def sorted_clue_keys(self):
        return sorted(self.clues.keys())

    def clues_by_type(self, clue_type):
        l = filter(lambda x: x.clue_type==clue_type, 
                   self.sorted_clue_keys)
        return [self.clues[x] for x in l]

    def get_clue(self, num, clue_type):
        k = ClueKey(num, clue_type)
        return self.clues[k]


import os
from setuptools import setup

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name='puzcw',
    version='0.0.8',
    author='Brian Wu',
    author_email='brian.george.wu@gmail.com',
    description='Python API for converting puz files to python objects',
    keywords=['crossword', 'puz'],
    packages=['puzcw'],
    install_requires=['pillow==3.3.0'],
    include_package_data = True,
    package_data={'static':['*']}
)